"""The Autowire decorator is used to
replace @api.marshal_with decorator and
@api.expect decorator through supporting MyPy types."""
import inspect
from functools import wraps
from flask_restx import Api      # type: ignore
from flask_fastx.model_api import create_model
from flask_fastx.parser_api import get_parser

# pylint: disable=C0207

class ApiDecorator():
    """The API Decorator"""

    def __init__(self) -> None:
        """Initialize the API decorator"""
        self.api = Api()

    def set_api(self, api: Api) -> None:
        """Set the API"""
        self.api = api

    def get_api(self) -> Api:
        """Get the API"""
        return self.api


API = ApiDecorator()


def register_api(api_main):
    """Register The Main App Api"""
    API.set_api(api_main)


def get_params_description(doc):
    """Get the parameters description from the docstring"""
    params_description = {}
    if doc is not None:
        doc = doc.split('\n')
        for line in doc:
            if ':' in line:
                line = line.split(':')
                line[0] = line[0].strip()
                params_description[line[0]] = line[1]
    return params_description


def autowire(func):
    """The Autowire Decorator that wraps the function"""
    api = API.get_api()
    model_name = func.__qualname__.split('.')[0]
    params_return = func.__annotations__
    params_return = params_return.get('return')
    isprimitive = False
    params_return_des = {}
    if params_return is None:
        api_model = api.model(model_name, {})
    else:
        if check_class_dict(params_return):
            params_return = {'data': params_return}
            isprimitive = True
        if hasattr(params_return, '__annotations__'):
            params_return_des = get_params_description(
                params_return.__doc__)
            params_return = params_return.__annotations__
        api_model = create_model(
            params_return, params_return_des)
        api_model = api.model(str(func.__name__)+model_name, api_model)
    signature = inspect.signature(func)
    parameters = dict(signature.parameters)
    parameters.pop('self')

    parser = get_parser(parameters)

    @wraps(func)
    @api.expect(parser)
    @api.marshal_with(api_model)
    def wrapper(*args, **kwargs):
        args_parser = parser.parse_args()
        if isprimitive:
            return {'data': func(*args, **args_parser, **kwargs)}
        return func(*args, **args_parser, **kwargs)
    return wrapper


def check_class_dict(param_type):
    """Check the class dict"""
    dict_type = str(param_type).replace("<", "")
    dict_type = dict_type.replace(">", "")
    dict_type = dict_type.split(" ")[-1]
    check_dect = dict_type != "dict" and not isinstance(param_type, dict)
    if not hasattr(param_type, '__annotations__') and check_dect:
        return True
    return False



'''
This files contains the implementation of the model generator function
it takes return type as input and returns a model as an output
'''
# pylint: disable=C0207

from flask_restx import fields   # type: ignore

# dict to map each return type to its field data type
python2restplus = {
    'str': fields.String,
    'datetime': fields.DateTime,
    'time': fields.String,
    'bool': fields.Boolean,
    'int': fields.Integer,
    'float': fields.Float,
    'relation': fields.Nested,
    'typing.List': fields.List,
    'dict': fields.Raw,
}


def create_model(types, descriptions):
    '''
    the model generator function which takes return type as input and returns a model as an output
    '''
    modeltemp = {}
    for type_ in types:
        ans = str(types[type_])

        if ans.find('typing') != -1:
            key = ans.split("[")[0]
            fields_param = ans.split("[")[1].replace("]", "")
            type_ = str(type_)
            if descriptions.get(type_) is None:
                descriptions[type_] = ""
            if python2restplus.get(fields_param) is None:
                fields_param = 'dict'
            modeltemp[type_] = python2restplus[key](
                python2restplus[fields_param], description=descriptions[type_])
        else:
            k = str(type_)
            ans = ans.split(" ")[1]
            ans = ans.replace("'", "")
            ans = ans.replace(">", "")
            if descriptions.get(k) is None:
                descriptions[k] = ""
            modeltemp[k] = python2restplus[ans](description=descriptions[k])
    return modeltemp



""" This module enable to add parameters to be parsed """

import ast
from flask_restx import reqparse  # type: ignore
from flask_fastx.params import Param

LocationToRestX = {
    'Query': 'args',
    'Body': 'json',
    'Form': 'form',
    'Header': 'headers',
    'Cookie': 'cookies',
    'File': 'files',
    'Path': 'path',
}


def get_parser(parameters) -> reqparse:
    """
    Parse all parameters which taken from the provided auto-wire decorator and
    return the results as a parser.
    """
    parser = reqparse.RequestParser()
    for param in parameters.values():
        location: str = get_param_location(param)
        param_type = get_param_type(param)
        default = None
        if location == 'path':
            continue
        param_default: Param
        param_default = param.default
        if param_default.default:
            default = param_default.default
        if str(param_type).find('typing.List') != -1:
            list_type = get_list_type(param_type)
            parser.add_argument(str(param.name), type=list_type, action='split',
                                location=location, default=default)

        elif str(param_type).find('typing.Literal') != -1:
            res = get_literal_tuple(str(param_type))
            if res and len(res) > 0:
                parser.add_argument(str(param.name), type=type(res[0]), choices=res,
                                    location=location, default=default)
        else:
            parser.add_argument(str(param.name), type=param_type,
                                location=location, default=default)

    return parser


def get_param_type(param):
    """ return parameter type """
    return param.annotation


def get_literal_tuple(param_type):
    """
    get literal option from annotation string and return result in tuple.
    """
    s_index: int = str(param_type).index('[')
    e_index: int = str(param_type).index(']')
    str_list = str(param_type)[s_index + 1:e_index]

    res = ast.literal_eval(str_list)
    if res:
        return tuple(res)

    return None


def get_param_location(param) -> str:
    """
    take a param and return his location in rest_x .
    """
    for location, value in LocationToRestX.items():
        if str(param).find(location) != -1:
            return value

    return LocationToRestX['Path']


def get_list_type(param_type):
    """
    return type of given list
    """
    if str(param_type).find('[str]') != -1:
        return str

    if str(param_type).find('[int]') != -1:
        return int

    if str(param_type).find('[float]') != -1:
        return float

    if str(param_type).find('[bool]') != -1:
        return bool

    return str



'''
This module is for extracting path params from a given route
'''

from typing import List
import re


def extract_path_params(path):
    '''
    Extracting path params from string by regex and append them to list
    '''
    params_types_list: List[str] = []  # Store path params with their type
    params_list: List[str] = []  # Store path params without type
    # Searching for params between angle brackets
    params = re.findall(r'\<(.*?)\>', path)  # returns a list with all matches
    if len(params) > 0:
        params_types_list.extend(params)
    # removing param type
    for param in params_types_list:
        param = param[param.find(':')+1:]
        params_list.append(param)
    return params_list



"""
define parameters location classes
"""

# pylint: disable=E0611
# pylint: disable=R0903

from enum import Enum
from typing import Any, Dict, Optional

from pydantic.fields import FieldInfo, Undefined  # type: ignore


class ParamTypes(Enum):
    """
    define ParamTypes
    """
    QUERY = "args"
    HEADER = "headers"
    PATH = "path"
    COOKIE = "cookies"
    BODY = 'json'
    FORM = 'form'
    FILE = 'files'





class Param(FieldInfo):
    """
    define Param class
    """
    in_: ParamTypes

    def __init__(
            self,
            default: Any,
            *,
            alias: Optional[str] = None,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            deprecated: Optional[bool] = None,
            **extra: Any,
    ):
        self.deprecated = deprecated
        self.example = example
        self.examples = examples
        super().__init__(
            default,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.default})"


class Path(Param):
    """
    define Path class
    """
    in_ = ParamTypes.PATH

    def __init__(
            self,
            default: Any,
            *,
            alias: Optional[str] = None,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            deprecated: Optional[bool] = None,
            **extra: Any,
    ):
        self.in_ = self.in_
        super().__init__(
            ...,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            deprecated=deprecated,
            example=example,
            examples=examples,
            **extra,
        )


class Query(Param):
    """
    define Query class
    """
    in_ = ParamTypes.QUERY

    def __init__(
            self,
            default: Any,
            *,
            alias: Optional[str] = None,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            deprecated: Optional[bool] = None,
            **extra: Any,
    ):
        self.in_ = self.in_
        super().__init__(
            default,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            deprecated=deprecated,
            example=example,
            examples=examples,
            **extra,
        )


class Header(Param):
    """
    define Header class
    """
    in_ = ParamTypes.HEADER

    def __init__(
            self,
            default: Any,
            *,
            alias: Optional[str] = None,
            convert_underscores: bool = True,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            deprecated: Optional[bool] = None,
            **extra: Any,
    ):
        self.convert_underscores = convert_underscores
        self.in_ = self.in_
        super().__init__(
            default,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            deprecated=deprecated,
            example=example,
            examples=examples,
            **extra,
        )


class Cookie(Param):
    """
    define Cookie class
    """
    in_ = ParamTypes.COOKIE

    def __init__(
            self,
            default: Any,
            *,
            alias: Optional[str] = None,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            deprecated: Optional[bool] = None,
            **extra: Any,
    ):
        self.in_ = self.in_
        super().__init__(
            default,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            deprecated=deprecated,
            example=example,
            examples=examples,
            **extra,
        )


class Body(Param):
    """
    define Body class
    """
    in_ = ParamTypes.BODY

    def __init__(
            self,
            default: Any,
            *,
            embed: bool = False,
            media_type: str = "application/json",
            alias: Optional[str] = None,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            **extra: Any,
    ):
        self.embed = embed
        self.media_type = media_type
        self.example = example
        self.examples = examples
        super().__init__(
            default,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.default})"


class Form(Param):
    """
    define Form class
    """
    in_ = ParamTypes.FORM

    def __init__(
            self,
            default: Any,
            *,
            media_type: str = "application/x-www-form-urlencoded",
            alias: Optional[str] = None,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            **extra: Any,
    ):
        super().__init__(
            default,
            embed=True,
            media_type=media_type,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            example=example,
            examples=examples,
            **extra,
        )


class File(Param):
    """
    define File class
    """
    in_ = ParamTypes.FILE

    def __init__(
            self,
            default: Any,
            *,
            media_type: str = "multipart/form-data",
            alias: Optional[str] = None,
            title: Optional[str] = None,
            description: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            regex: Optional[str] = None,
            example: Any = Undefined,
            examples: Optional[Dict[str, Any]] = None,
            **extra: Any,
    ):
        super().__init__(
            default,
            media_type=media_type,
            alias=alias,
            title=title,
            description=description,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            example=example,
            examples=examples,
            **extra,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flask_fastx']

package_data = \
{'': ['*']}

install_requires = \
['flask-restx>=0.5.0,<0.6.0', 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'flask-fastx',
    'version': '0.1.0',
    'description': 'Flask-Fastx is a Fast API style support for Flask. It Gives you MyPy types with the flexibility of flask.',
    'long_description': "==============\nFlask-Fastx\n==============\n\n.. image:: https://codecov.io/gh/tactful-ai/flask-faster-api/branch/main/graph/badge.svg?token=FZJANN69LH\n    :target: https://codecov.io/gh/tactful-ai/flask-faster-api\n    :alt: Code Coverage\n    \n.. image:: https://github.com/tactful-ai/flask-faster-api/actions/workflows/python-package.yml/badge.svg\n   :target: https://github.com/tactful-ai/flask-faster-api/actions/workflows/python-package.yml\n   :alt: CI Tests\n.. image:: https://img.shields.io/github/license/tactful-ai/flask-faster-api   \n    :alt: License\n.. image:: https://img.shields.io/github/stars/tactful-ai/flask-faster-api?style=social   :alt: GitHub Repo stars\n.. image:: https://img.shields.io/pypi/pyversions/flask-restx-square  \n    :target: https://pypi.org/project/flask-fastx\n    :alt: Supported Python versions\n   \n\nFlask-Fastx is a Fast API style support for Flask. It Gives you MyPy types with the flexibility of flask.\n\n\n\nCompatibility\n=============\n\nFlask-Fastx requires Python 3.7+. \n\n\n\nInstallation\n============\n\nYou can install Flask-Fastx with pip:\n\n.. code-block:: console\n\n    $ pip install flask-fastx\n    \n\nQuick start\n===========\n\nWith Flask-Fastx and the autowire decorator feature, parsing path parameters and creating api models is done using only one decorator! \n\nBefore Using Flask-Fastx\n================================\n\n.. image:: https://user-images.githubusercontent.com/63073172/130029969-2f59777d-3ad1-423f-bbf1-a8c73d056300.png\n\n\nAfter Using Flask-Fastx\n================================\n\n.. image:: https://user-images.githubusercontent.com/63073172/130017489-55b0d09d-5ba5-4455-8c88-29878b39e6c5.png\n\n\n\n\nContributors\n============\n\nFlask-Fastx is brought to you by @mfouad, @seifashraf1, @ahmedihabb2, @nadaabdelmaboud, @omargamal253\n\n\n\n\nContribution\n============\nWant to contribute? That's awesome! (Details Soon) \n",
    'author': 'nadaabdelmaboud     seifashraf1     ahmedihabb2     omargamal253',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tactful-ai/flask-faster-api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
